import os
import boto3

# Setup HCP bucket
access_key = open('/storage/gold/cinn/2013/neuroade/Thalamus/Nipype/.creds/aws_access_key', 'r')
access_key = access_key.readlines()
access_key = str(access_key[0][:-1])
private_key = open('/storage/gold/cinn/2013/neuroade/Thalamus/Nipype/.creds/aws_private_key', 'r')
private_key = private_key.readlines()
private_key = str(private_key[0][:-1])

hcp = boto3.resource(service_name='s3', region_name='us-east-1', aws_access_key_id=access_key, aws_secret_access_key=private_key)
hcp = hcp.Bucket('hcp-openaccess')

'''
get_data takes an ID for a single subject as input.
It downloads the files for that subject from the S3 bucket to the directory data_dir
A dictionary containing paths to files is returned as output.
'''

def get_data(subject, data_dir, section='default'):
    
    import os
        
    # misc
    if not os.path.isfile(f'{data_dir}/{subject}_3T.csv'):
        hcp.download_file(f'HCP_1200/{subject}/unprocessed/3T/{subject}_3T.csv', f'{data_dir}/{subject}_3T.csv')
        
    if section == 'default' or section == 'anat':
        # Download T1w images
        if not os.path.isfile(f'{data_dir}/{subject}_T1w_acpc_dc_restore_1.25.nii.gz'):
            hcp.download_file(f'HCP_1200/{subject}/T1w/T1w_acpc_dc_restore_1.25.nii.gz', f'{data_dir}/{subject}_T1w_acpc_dc_restore_1.25.nii.gz')
        if not os.path.isfile(f'{data_dir}/{subject}_T1w_acpc_dc_restore.nii.gz'):
            hcp.download_file(f'HCP_1200/{subject}/T1w/T1w_acpc_dc_restore.nii.gz', f'{data_dir}/{subject}_T1w_acpc_dc_restore.nii.gz')
        if not os.path.isfile(f'{data_dir}/{subject}_T1w_acpc_dc_restore_brain.nii.gz'):
            hcp.download_file(f'HCP_1200/{subject}/T1w/T1w_acpc_dc_restore_brain.nii.gz', f'{data_dir}/{subject}_T1w_acpc_dc_restore_brain.nii.gz')

    if section == 'default' or section == 'anat':
        # Download Diffusion images   
        if not os.path.isfile(f'{data_dir}/{subject}_Diffusion_bvals'):
            hcp.download_file(f'HCP_1200/{subject}/T1w/Diffusion/bvals', f'{data_dir}/{subject}_Diffusion_bvals')
        if not os.path.isfile(f'{data_dir}/{subject}_Diffusion_bvecs'):
            hcp.download_file(f'HCP_1200/{subject}/T1w/Diffusion/bvecs', f'{data_dir}/{subject}_Diffusion_bvecs')    
        if not os.path.isfile(f'{data_dir}/{subject}_Diffusion_data.nii.gz'):
            hcp.download_file(f'HCP_1200/{subject}/T1w/Diffusion/data.nii.gz', f'{data_dir}/{subject}_Diffusion_data.nii.gz') 
        if not os.path.isfile(f'{data_dir}/{subject}_Diffusion_nodif_brain_mask.nii.gz'):
            hcp.download_file(f'HCP_1200/{subject}/T1w/Diffusion/nodif_brain_mask.nii.gz', f'{data_dir}/{subject}_Diffusion_nodif_brain_mask.nii.gz')
    
    # Create dictionary with file paths
    data_dict = {
            'csv': f'{data_dir}/{subject}_3T.csv',
            'T1w': f'{data_dir}/{subject}_T1w_acpc_dc_restore.nii.gz',
            'T1w_125': f'{data_dir}/{subject}_T1w_acpc_dc_restore_1.25.nii.gz',
            'T1w_brain': f'{data_dir}/{subject}_T1w_acpc_dc_restore_brain.nii.gz',
            'DWI_bvals': f'{data_dir}/{subject}_Diffusion_bvals',
            'DWI_bvecs': f'{data_dir}/{subject}_Diffusion_bvecs',
            'DWI': f'{data_dir}/{subject}_Diffusion_data.nii.gz',
            'DWI_mask': f'{data_dir}/{subject}_Diffusion_nodif_brain_mask.nii.gz'
    }
 
    
    return data_dict

def get_edge(in_file):
    import mahotas
    import nibabel as nib

    img = nib.load(in_file)
    data = img.get_fdata()
    edge = mahotas.labeled.bwperim(data) 
    output = nib.Nifti1Image(edge, header = img.header, affine = img.affine)
    if in_file.endswith('.nii.gz'):
        in_file=in_file[:-7]
        out_file=f'{in_file}-mahotas.nii.gz'
    elif in_file.endswith('.nii'):
        in_file=in_file[:-4]
        out_file=f'{in_file}-mahotas.nii'
    nib.nifti1.save(output, out_file)
    return out_file

def return_n(in_file, n):
    out_file = in_file[n]
    return out_file
    
# FSL-qboot
from nipype.interfaces.fsl.base import (FSLCommand, FSLCommandInputSpec)
from nipype.interfaces.freesurfer.base import FSCommand
from nipype.interfaces.base import (File, TraitedSpec, traits, Directory, CommandLineInputSpec, BaseInterface, BaseInterfaceInputSpec)
from nipype.interfaces.io import FreeSurferSource
from nipype.interfaces.matlab import MatlabCommand, MatlabInputSpec
from nipype import Function
from string import Template
import os

class CustomQBOOTInputSpec(FSLCommandInputSpec):
    data = File(exists=True, mandatory=True, argstr='--data=%s', position=0, desc='Data file')
    mask = File(exists=True, mandatory=True, argstr='--mask=%s', position=1, desc='Mask file')
    bvecs = File(exists=True, mandatory=True, argstr='--bvecs=%s', position=2, desc='b vectors file')
    bvals = File(exists=True, mandatory=True, argstr='--bvals=%s', position=3, desc='b values file')
    shbasis = traits.Int(mandatory=True, argstr='--lmax=%s', desc='Maximum spherical harmonic order employed (must be even, default=4)')
    savemeancoeff = traits.Bool(desc='Save the mean ODF coefficients across all samples', argstr='--savemeancoeff')

class CustomQBOOTOutputSpec(TraitedSpec):
    out_file = File(desc='the output image')


class CustomQBOOT(FSLCommand):
    _cmd = 'qboot'
    input_spec = CustomQBOOTInputSpec
    output_spec = CustomQBOOTOutputSpec
    
    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_file'] = os.path.abspath('./logdir/mean_SHcoeff.nii.gz')
        return outputs
    
# Freesurfer thalamic segmentation
class CustomFSThalamusSegInputSpec(CommandLineInputSpec):
    subject_id = traits.Str(argstr='%s', desc='subject name', position=0, mandatory=True)
    subjects_dir = Directory(exists=True, argstr='%s', hash_files=False, desc='path to subjects directory', genfile=True)

class CustomFSThalamusSegOutputSpec(TraitedSpec):
    subjects_dir = Directory(desc="Freesurfer subjects directory.")
    subject_id = traits.Str(desc='Subject name')
    tv_txt = File(desc='estimated volumes of thalamic nuclei')
    tv_sub = File(desc='segmentation at subvoxel resolution (0.5mm)')
    tv = File(desc='segmentation in freesurfer voxel space (1mm, unless -cm flag is used)')
    
class CustomFSThalamusSeg(FSCommand):
    _cmd = 'segmentThalamicNuclei.sh'
    input_spec = CustomFSThalamusSegInputSpec
    output_spec = CustomFSThalamusSegOutputSpec
    
    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['subject_id'] = self.inputs.subject_id
        outputs['subjects_dir'] = self.inputs.subjects_dir
        outputs['tv_txt'] = f'/storage/gold/cinn/2013/neuroade/Thalamus/temp/Data/main/reconall_PARTICIPANTID/{subject}/mri/ThalamicNuclei.v12.T1.volumes.txt'
        outputs['tv_sub'] = f'/storage/gold/cinn/2013/neuroade/Thalamus/temp/Data/main/reconall_PARTICIPANTID/{subject}/mri/ThalamicNuclei.v12.T1.mgz'
        outputs['tv'] = f'/storage/gold/cinn/2013/neuroade/Thalamus/temp/Data/main/reconall_PARTICIPANTID/{subject}/mri/ThalamicNuclei.v12.T1.FSvoxelSpace.mgz'
        return outputs

# Get participant data
subject=PARTICIPANTID
data_dir='/storage/gold/cinn/2013/neuroade/Thalamus/temp/Data'
data = get_data(subject, data_dir)

# Import packages
import os
from nipype.interfaces.slicer import (GaussianBlurImageFilter, BRAINSFit)
from nipype.interfaces.fsl import (Eddy, FLIRT, DTIFit, Merge, FAST, BET, ApplyWarp, ConvertXFM, 
                                    DilateImage, BinaryMaths, Threshold)
from nipype.interfaces.spm import Segment
from nipype.interfaces.freesurfer import (ReconAll, Label2Vol, Binarize, MRIConvert)
from nipype.interfaces.io import DataSink
import nipype.interfaces.matlab as matlab
from nipype.interfaces.utility import Merge as Mergefiles
from nipype import (Node, Workflow, MapNode)

mni_template='/storage/gold/cinn/2013/neuroade/Thalamus/temp/Install/fsl/data/standard/MNI152_T1_1mm_brain.nii.gz'
data_out=f'/storage/gold/cinn/2013/neuroade/Thalamus/Output/{subject}'

if not os.path.exists(data_out):
    os.makedirs(data_out)

datasink = Node(DataSink(base_directory=data_out), name='datasink_PARTICIPANTID')

fast = Node(FAST(img_type=1), name='FAST_T1_PARTICIPANTID')
get_csfprob = Node(Function(input_names=['in_file','n'], output_names=['out_file'], function=return_n), name='Get_CSF_PartialVolume_PARTICIPANTID')
get_csfprob.inputs.n = 0
csf_thr = Node(Threshold(thresh=0.05, direction='below', args='-bin'), name='Threshold_lessthan005_PARTICIPANTID')
csf_mask = Node(BinaryMaths(operation='sub', in_file=data['DWI_mask']), name='csf_mask_PARTICIPANTID')
anat2mni = Node(FLIRT(reference=mni_template, in_file=data['T1w'], dof=12), name='FLIRT_anat2mni_PARTICIPANTID')
dwi2mni = Node(FLIRT(reference=mni_template, in_file=data['T1w_125'], dof=12), name='FLIRT_dwi2mni_PARTICIPANTID')
reconall = Node(ReconAll(T1_files=data['T1w'], subject_id=str(subject), flags='-cm'), name='reconall_PARTICIPANTID')
fsThal = Node(CustomFSThalamusSeg(),name='Thalamus_seg_FS_PARTICIPANTID')
label2vol = Node(Label2Vol(), name='Conv_thalSeg2anat_space_PARTICIPANTID')
brain125 = Node(BinaryMaths(operation='mul', in_file=data['T1w_125'], operand_file=data['DWI_mask']),name='Brain_diffspace_PARTICIPANTID')
br2br125 = Node(FLIRT(in_file=data['T1w_brain'], dof=7),name='FLIRT_brain2brain125_PARTICIPANTID')
flirt_lThal = Node(FLIRT(apply_xfm=True), name='FLIRT_lThal_PARTICIPANTID')
flirt_rThal = Node(FLIRT(apply_xfm=True), name='FLIRT_rThal_PARTICIPANTID')
lThal_bin = Node(Threshold(thresh=0.5, direction='below', args='-bin'), name='Threshold_lThal_PARTICIPANTID')
rThal_bin = Node(Threshold(thresh=0.5, direction='below', args='-bin'), name='Threshold_rThal_PARTICIPANTID')
label2vol_dwi = Node(Label2Vol(), name='Conv_aseg2anat_space_PARTICIPANTID')
binarize_dwi_l = Node(Binarize(match=[10]),name='Get_wholeThal_l_PARTICIPANTID')
mriconvert = Node(MRIConvert(out_type='niigz'), name='Convert2NII_FSseg_PARTICIPANTID')
mriconvert_dwi_l = Node(MRIConvert(out_type='niigz'), name='Convert2NII_wholeThal_l_PARTICIPANTID')
dtifit = Node(DTIFit(bvals=data['DWI_bvals'], bvecs=data['DWI_bvecs'], dwi=data['DWI'], mask=data['DWI_mask']), name='GetFA_PARTICIPANTID')
getedge_l = Node(Function(input_names=['in_file'], output_names=['out_file'], function=get_edge), name='FindPerimeter_l_PARTICIPANTID')
dilate_l = Node(DilateImage(operation='mean'), name='DilatePerimeter_l_PARTICIPANTID')
edgearea_l = Node(BinaryMaths(operation='mul'), name='RefinePerimeter_l_PARTICIPANTID')
fa_thr = Node(Threshold(thresh=0.55, direction='below', args='-bin'), name='Threshold_greaterthan055_PARTICIPANTID') 
fa_area_l = Node(BinaryMaths(operation='mul'), name='FA_exclusions_l_PARTICIPANTID')
thal_neg_fa_l = Node(BinaryMaths(operation='sub'), name='Thal_FA_corr_l_PARTICIPANTID')
thal_csf_fa_cor_l = Node(BinaryMaths(operation='mul'), name='Thal_CSF_FA_corr_l_PARTICIPANTID')
binarize_dwi_r = Node(Binarize(match=[49]),name='Get_wholeThal_r_PARTICIPANTID')
mriconvert_dwi_r = Node(MRIConvert(out_type='niigz'), name='Convert2NII_wholeThal_r_PARTICIPANTID')
getedge_r = Node(Function(input_names=['in_file'], output_names=['out_file'], function=get_edge), name='FindPerimeter_r_PARTICIPANTID')
dilate_r = Node(DilateImage(operation='mean'), name='DilatePerimeter_r_PARTICIPANTID')
edgearea_r = Node(BinaryMaths(operation='mul'), name='RefinePerimeter_r_PARTICIPANTID')
fa_area_r = Node(BinaryMaths(operation='mul'), name='FA_exclusions_r_PARTICIPANTID')
thal_neg_fa_r = Node(BinaryMaths(operation='sub'), name='Thal_FA_corr_r_PARTICIPANTID')
thal_csf_fa_cor_r = Node(BinaryMaths(operation='mul'), name='Thal_CSF_FA_corr_r_PARTICIPANTID')
qboot = Node(CustomQBOOT(data=data['DWI'], mask=data['DWI_mask'], bvals=data['DWI_bvals'], bvecs=data['DWI_bvecs'], shbasis=6, savemeancoeff=True), name='qboot_PARTICIPANTID')

wf = Workflow(name='main', base_dir='/storage/gold/cinn/2013/neuroade/Thalamus/temp/Data')

wf.add_nodes([anat2mni])
wf.connect([(reconall, fsThal, [('subject_id','subject_id')]),
            (reconall, fsThal, [('subjects_dir','subjects_dir')]),
            (fast, get_csfprob, [('partial_volume_files','in_file')]),
            (get_csfprob, csf_thr, [('out_file','in_file')]),
            (fsThal, label2vol, [('tv','seg_file')]),
            (fsThal, label2vol, [('tv','reg_header')]),
            (reconall, label2vol, [('rawavg','template_file')]),
            (reconall, label2vol_dwi, [('aseg','seg_file')]),
            (reconall, label2vol_dwi, [('aseg','reg_header')]),
            (reconall, label2vol_dwi, [('rawavg','template_file')]),
            (label2vol, mriconvert, [('vol_label_file','in_file')]),
            (label2vol_dwi, binarize_dwi_l, [('vol_label_file','in_file')]),
            (label2vol_dwi, binarize_dwi_r, [('vol_label_file','in_file')]),
            (thal_csf_fa_cor_l, datasink, [('out_file','lthal_dwi')]),
            (thal_csf_fa_cor_r, datasink, [('out_file','rthal_dwi')]),
            (qboot, datasink, [('out_file','qboot_ODFs')]),
            (binarize_dwi_l, mriconvert_dwi_l, [('binary_file','in_file')]),
            (binarize_dwi_r, mriconvert_dwi_r, [('binary_file','in_file')]),
            (mriconvert, datasink, [('out_file','anatSeg')]),
            (lThal_bin, getedge_l, [('out_file','in_file')]),
            (rThal_bin, getedge_r, [('out_file','in_file')]),
            (getedge_l, dilate_l, [('out_file','in_file')]),
            (getedge_r, dilate_r, [('out_file','in_file')]),
            (lThal_bin, edgearea_l, [('out_file','operand_file')]),
            (rThal_bin, edgearea_r, [('out_file','operand_file')]),
            (dilate_l, edgearea_l, [('out_file','in_file')]),
            (dilate_r, edgearea_r, [('out_file','in_file')]),
            (dtifit, fa_thr, [('FA','in_file')]),
            (fa_thr, fa_area_l, [('out_file','in_file')]),
            (fa_thr, fa_area_r, [('out_file','in_file')]),
            (edgearea_l, fa_area_l, [('out_file','operand_file')]),
            (edgearea_r, fa_area_r, [('out_file','operand_file')]),
            (lThal_bin, thal_neg_fa_l, [('out_file','in_file')]),
            (rThal_bin, thal_neg_fa_r, [('out_file','in_file')]),
            (fa_area_l, thal_neg_fa_l, [('out_file','operand_file')]),
            (fa_area_r, thal_neg_fa_r, [('out_file','operand_file')]),
            (thal_neg_fa_l, thal_csf_fa_cor_l, [('out_file','in_file')]),
            (thal_neg_fa_r, thal_csf_fa_cor_r, [('out_file','in_file')]),
            (csf_mask, thal_csf_fa_cor_l, [('out_file','operand_file')]),
            (csf_mask, thal_csf_fa_cor_r, [('out_file','operand_file')]),
            (brain125, br2br125,[('out_file','reference')]),
            (brain125, flirt_lThal, [('out_file','reference')]),
            (brain125, flirt_rThal, [('out_file','reference')]),
            (mriconvert_dwi_l, flirt_lThal, [('out_file','in_file')]),
            (mriconvert_dwi_r, flirt_rThal, [('out_file','in_file')]),
            (br2br125, flirt_lThal, [('out_matrix_file','in_matrix_file')]),
            (br2br125, flirt_rThal, [('out_matrix_file','in_matrix_file')]),
            (flirt_lThal, lThal_bin, [('out_file','in_file')]),
            (flirt_rThal, rThal_bin, [('out_file','in_file')]),
            (brain125, fast, [('out_file','in_files')]),
            (csf_thr, csf_mask, [('out_file','operand_file')]),
            (dwi2mni, datasink, [('out_matrix_file','dwi2mni_1mm_warp')]),
            (anat2mni, datasink, [('out_matrix_file','anat2mni_1mm_warp')])])

wf.run()

#Remove delete all files from storage drive
import glob
for file in glob.glob(f'/storage/gold/cinn/2013/neuroade/Thalamus/temp/Data/{subject}_*'):
    os.remove(file)
